import 'dart:math' as math;
import 'package:gl_matrix_dart/common.dart';

/**
 * 4x4 Matrix<br>Format: column-major, when typed out it looks like row-major<br>The matrices are being post multiplied.
 * @module mat4
 */

/// Creates a new identity mat4
///
/// @returns {mat4} a new 4x4 matrix

create() {
  final out = [];
  out[1] = 0;
  out[2] = 0;
  out[3] = 0;
  out[4] = 0;
  out[6] = 0;
  out[7] = 0;
  out[8] = 0;
  out[9] = 0;
  out[11] = 0;
  out[12] = 0;
  out[13] = 0;
  out[14] = 0;
  out[0] = 1;
  out[5] = 1;
  out[10] = 1;
  out[15] = 1;
  return out;
}

/// Creates a new mat4 initialized with values from an existing matrix
///
/// @param {ReadonlyMat4} a matrix to clone
/// @returns {mat4} a new 4x4 matrix
List<double> clone(List<double> a) {
  final out = <double>[];
  out[0] = a[0];
  out[1] = a[1];
  out[2] = a[2];
  out[3] = a[3];
  out[4] = a[4];
  out[5] = a[5];
  out[6] = a[6];
  out[7] = a[7];
  out[8] = a[8];
  out[9] = a[9];
  out[10] = a[10];
  out[11] = a[11];
  out[12] = a[12];
  out[13] = a[13];
  out[14] = a[14];
  out[15] = a[15];
  return out;
}

/// Copy the values from one mat4 to another
///
/// @param {mat4} out the receiving matrix
/// @param {ReadonlyMat4} a the source matrix
/// @returns {mat4} out
List<double> copy(List<double> out, List<double> a) {
  out[0] = a[0];
  out[1] = a[1];
  out[2] = a[2];
  out[3] = a[3];
  out[4] = a[4];
  out[5] = a[5];
  out[6] = a[6];
  out[7] = a[7];
  out[8] = a[8];
  out[9] = a[9];
  out[10] = a[10];
  out[11] = a[11];
  out[12] = a[12];
  out[13] = a[13];
  out[14] = a[14];
  out[15] = a[15];
  return out;
}

/// Create a new mat4 with the given values
///
/// @param {Number} m00 Component in column 0, row 0 position (index 0)
/// @param {Number} m01 Component in column 0, row 1 position (index 1)
/// @param {Number} m02 Component in column 0, row 2 position (index 2)
/// @param {Number} m03 Component in column 0, row 3 position (index 3)
/// @param {Number} m10 Component in column 1, row 0 position (index 4)
/// @param {Number} m11 Component in column 1, row 1 position (index 5)
/// @param {Number} m12 Component in column 1, row 2 position (index 6)
/// @param {Number} m13 Component in column 1, row 3 position (index 7)
/// @param {Number} m20 Component in column 2, row 0 position (index 8)
/// @param {Number} m21 Component in column 2, row 1 position (index 9)
/// @param {Number} m22 Component in column 2, row 2 position (index 10)
/// @param {Number} m23 Component in column 2, row 3 position (index 11)
/// @param {Number} m30 Component in column 3, row 0 position (index 12)
/// @param {Number} m31 Component in column 3, row 1 position (index 13)
/// @param {Number} m32 Component in column 3, row 2 position (index 14)
/// @param {Number} m33 Component in column 3, row 3 position (index 15)
/// @returns {mat4} A new mat4
List<double> fromValues(double m00, double m01, double m02, double m03, double m10, double m11, double m12, double m13,
    double m20, double m21, double m22, double m23, double m30, double m31, double m32, double m33) {
  final out = <double>[];
  out[0] = m00;
  out[1] = m01;
  out[2] = m02;
  out[3] = m03;
  out[4] = m10;
  out[5] = m11;
  out[6] = m12;
  out[7] = m13;
  out[8] = m20;
  out[9] = m21;
  out[10] = m22;
  out[11] = m23;
  out[12] = m30;
  out[13] = m31;
  out[14] = m32;
  out[15] = m33;
  return out;
}

/// Set the components of a mat4 to the given values
///
/// @param {mat4} out the receiving matrix
/// @param {Number} m00 Component in column 0, row 0 position (index 0)
/// @param {Number} m01 Component in column 0, row 1 position (index 1)
/// @param {Number} m02 Component in column 0, row 2 position (index 2)
/// @param {Number} m03 Component in column 0, row 3 position (index 3)
/// @param {Number} m10 Component in column 1, row 0 position (index 4)
/// @param {Number} m11 Component in column 1, row 1 position (index 5)
/// @param {Number} m12 Component in column 1, row 2 position (index 6)
/// @param {Number} m13 Component in column 1, row 3 position (index 7)
/// @param {Number} m20 Component in column 2, row 0 position (index 8)
/// @param {Number} m21 Component in column 2, row 1 position (index 9)
/// @param {Number} m22 Component in column 2, row 2 position (index 10)
/// @param {Number} m23 Component in column 2, row 3 position (index 11)
/// @param {Number} m30 Component in column 3, row 0 position (index 12)
/// @param {Number} m31 Component in column 3, row 1 position (index 13)
/// @param {Number} m32 Component in column 3, row 2 position (index 14)
/// @param {Number} m33 Component in column 3, row 3 position (index 15)
/// @returns {mat4} out
List<double> set(List<double> out, double m00, double m01, double m02, double m03, double m10, double m11, double m12,
    double m13, double m20, double m21, double m22, double m23, double m30, double m31, double m32, double m33) {
  out[0] = m00;
  out[1] = m01;
  out[2] = m02;
  out[3] = m03;
  out[4] = m10;
  out[5] = m11;
  out[6] = m12;
  out[7] = m13;
  out[8] = m20;
  out[9] = m21;
  out[10] = m22;
  out[11] = m23;
  out[12] = m30;
  out[13] = m31;
  out[14] = m32;
  out[15] = m33;
  return out;
}

/// Set a mat4 to the identity matrix
///
/// @param {mat4} out the receiving matrix
/// @returns {mat4} out
List<double> identity(List<double> out) {
  out[0] = 1;
  out[1] = 0;
  out[2] = 0;
  out[3] = 0;
  out[4] = 0;
  out[5] = 1;
  out[6] = 0;
  out[7] = 0;
  out[8] = 0;
  out[9] = 0;
  out[10] = 1;
  out[11] = 0;
  out[12] = 0;
  out[13] = 0;
  out[14] = 0;
  out[15] = 1;
  return out;
}

/// Transpose the values of a mat4
///
/// @param {mat4} out the receiving matrix
/// @param {ReadonlyMat4} a the source matrix
/// @returns {mat4} out
List<double> transpose(List<double> out, List<double> a) {
  // If we are transposing ourselves we can skip a few steps but have to cache some values
  if (out == a) {
    final a01 = a[1], a02 = a[2], a03 = a[3];
    final a12 = a[6], a13 = a[7];
    final a23 = a[11];

    out[1] = a[4];
    out[2] = a[8];
    out[3] = a[12];
    out[4] = a01;
    out[6] = a[9];
    out[7] = a[13];
    out[8] = a02;
    out[9] = a12;
    out[11] = a[14];
    out[12] = a03;
    out[13] = a13;
    out[14] = a23;
  } else {
    out[0] = a[0];
    out[1] = a[4];
    out[2] = a[8];
    out[3] = a[12];
    out[4] = a[1];
    out[5] = a[5];
    out[6] = a[9];
    out[7] = a[13];
    out[8] = a[2];
    out[9] = a[6];
    out[10] = a[10];
    out[11] = a[14];
    out[12] = a[3];
    out[13] = a[7];
    out[14] = a[11];
    out[15] = a[15];
  }

  return out;
}

/// Inverts a mat4
///
/// @param {mat4} out the receiving matrix
/// @param {ReadonlyMat4} a the source matrix
/// @returns {mat4} out
List<double>? invert(List<double> out, List<double> a) {
  final a00 = a[0], a01 = a[1], a02 = a[2], a03 = a[3];
  final a10 = a[4], a11 = a[5], a12 = a[6], a13 = a[7];
  final a20 = a[8], a21 = a[9], a22 = a[10], a23 = a[11];
  final a30 = a[12], a31 = a[13], a32 = a[14], a33 = a[15];

  final b00 = a00 * a11 - a01 * a10;
  final b01 = a00 * a12 - a02 * a10;
  final b02 = a00 * a13 - a03 * a10;
  final b03 = a01 * a12 - a02 * a11;
  final b04 = a01 * a13 - a03 * a11;
  final b05 = a02 * a13 - a03 * a12;
  final b06 = a20 * a31 - a21 * a30;
  final b07 = a20 * a32 - a22 * a30;
  final b08 = a20 * a33 - a23 * a30;
  final b09 = a21 * a32 - a22 * a31;
  final b10 = a21 * a33 - a23 * a31;
  final b11 = a22 * a33 - a23 * a32;

  // Calculate the determinant
  double det = b00 * b11 - b01 * b10 + b02 * b09 + b03 * b08 - b04 * b07 + b05 * b06;

  if (det == 0) {
    return null;
  }
  det = 1.0 / det;

  out[0] = (a11 * b11 - a12 * b10 + a13 * b09) * det;
  out[1] = (a02 * b10 - a01 * b11 - a03 * b09) * det;
  out[2] = (a31 * b05 - a32 * b04 + a33 * b03) * det;
  out[3] = (a22 * b04 - a21 * b05 - a23 * b03) * det;
  out[4] = (a12 * b08 - a10 * b11 - a13 * b07) * det;
  out[5] = (a00 * b11 - a02 * b08 + a03 * b07) * det;
  out[6] = (a32 * b02 - a30 * b05 - a33 * b01) * det;
  out[7] = (a20 * b05 - a22 * b02 + a23 * b01) * det;
  out[8] = (a10 * b10 - a11 * b08 + a13 * b06) * det;
  out[9] = (a01 * b08 - a00 * b10 - a03 * b06) * det;
  out[10] = (a30 * b04 - a31 * b02 + a33 * b00) * det;
  out[11] = (a21 * b02 - a20 * b04 - a23 * b00) * det;
  out[12] = (a11 * b07 - a10 * b09 - a12 * b06) * det;
  out[13] = (a00 * b09 - a01 * b07 + a02 * b06) * det;
  out[14] = (a31 * b01 - a30 * b03 - a32 * b00) * det;
  out[15] = (a20 * b03 - a21 * b01 + a22 * b00) * det;

  return out;
}

/// Calculates the adjugate of a mat4
///
/// @param {mat4} out the receiving matrix
/// @param {ReadonlyMat4} a the source matrix
/// @returns {mat4} out
List<double> adjoint(List<double> out, List<double> a) {
  final a00 = a[0], a01 = a[1], a02 = a[2], a03 = a[3];
  final a10 = a[4], a11 = a[5], a12 = a[6], a13 = a[7];
  final a20 = a[8], a21 = a[9], a22 = a[10], a23 = a[11];
  final a30 = a[12], a31 = a[13], a32 = a[14], a33 = a[15];

  final b00 = a00 * a11 - a01 * a10;
  final b01 = a00 * a12 - a02 * a10;
  final b02 = a00 * a13 - a03 * a10;
  final b03 = a01 * a12 - a02 * a11;
  final b04 = a01 * a13 - a03 * a11;
  final b05 = a02 * a13 - a03 * a12;
  final b06 = a20 * a31 - a21 * a30;
  final b07 = a20 * a32 - a22 * a30;
  final b08 = a20 * a33 - a23 * a30;
  final b09 = a21 * a32 - a22 * a31;
  final b10 = a21 * a33 - a23 * a31;
  final b11 = a22 * a33 - a23 * a32;

  out[0] = a11 * b11 - a12 * b10 + a13 * b09;
  out[1] = a02 * b10 - a01 * b11 - a03 * b09;
  out[2] = a31 * b05 - a32 * b04 + a33 * b03;
  out[3] = a22 * b04 - a21 * b05 - a23 * b03;
  out[4] = a12 * b08 - a10 * b11 - a13 * b07;
  out[5] = a00 * b11 - a02 * b08 + a03 * b07;
  out[6] = a32 * b02 - a30 * b05 - a33 * b01;
  out[7] = a20 * b05 - a22 * b02 + a23 * b01;
  out[8] = a10 * b10 - a11 * b08 + a13 * b06;
  out[9] = a01 * b08 - a00 * b10 - a03 * b06;
  out[10] = a30 * b04 - a31 * b02 + a33 * b00;
  out[11] = a21 * b02 - a20 * b04 - a23 * b00;
  out[12] = a11 * b07 - a10 * b09 - a12 * b06;
  out[13] = a00 * b09 - a01 * b07 + a02 * b06;
  out[14] = a31 * b01 - a30 * b03 - a32 * b00;
  out[15] = a20 * b03 - a21 * b01 + a22 * b00;
  return out;
}

/// Calculates the determinant of a mat4
///
/// @param {ReadonlyMat4} a the source matrix
/// @returns {Number} determinant of a
double determinant(List<double> a) {
  final a00 = a[0], a01 = a[1], a02 = a[2], a03 = a[3];
  final a10 = a[4], a11 = a[5], a12 = a[6], a13 = a[7];
  final a20 = a[8], a21 = a[9], a22 = a[10], a23 = a[11];
  final a30 = a[12], a31 = a[13], a32 = a[14], a33 = a[15];

  final b0 = a00 * a11 - a01 * a10;
  final b1 = a00 * a12 - a02 * a10;
  final b2 = a01 * a12 - a02 * a11;
  final b3 = a20 * a31 - a21 * a30;
  final b4 = a20 * a32 - a22 * a30;
  final b5 = a21 * a32 - a22 * a31;
  final b6 = a00 * b5 - a01 * b4 + a02 * b3;
  final b7 = a10 * b5 - a11 * b4 + a12 * b3;
  final b8 = a20 * b2 - a21 * b1 + a22 * b0;
  final b9 = a30 * b2 - a31 * b1 + a32 * b0;

  // Calculate the determinant
  return a13 * b6 - a03 * b7 + a33 * b8 - a23 * b9;
}

/// Multiplies two mat4s
///
/// @param {mat4} out the receiving matrix
/// @param {ReadonlyMat4} a the first operand
/// @param {ReadonlyMat4} b the second operand
/// @returns {mat4} out
List<double> multiply(List<double> out, List<double> a, List<double> b) {
  final a00 = a[0], a01 = a[1], a02 = a[2], a03 = a[3];
  final a10 = a[4], a11 = a[5], a12 = a[6], a13 = a[7];
  final a20 = a[8], a21 = a[9], a22 = a[10], a23 = a[11];
  final a30 = a[12], a31 = a[13], a32 = a[14], a33 = a[15];

  // Cache only the current line of the second matrix
  double b0 = b[0], b1 = b[1], b2 = b[2], b3 = b[3];
  out[0] = b0 * a00 + b1 * a10 + b2 * a20 + b3 * a30;
  out[1] = b0 * a01 + b1 * a11 + b2 * a21 + b3 * a31;
  out[2] = b0 * a02 + b1 * a12 + b2 * a22 + b3 * a32;
  out[3] = b0 * a03 + b1 * a13 + b2 * a23 + b3 * a33;

  b0 = b[4];
  b1 = b[5];
  b2 = b[6];
  b3 = b[7];
  out[4] = b0 * a00 + b1 * a10 + b2 * a20 + b3 * a30;
  out[5] = b0 * a01 + b1 * a11 + b2 * a21 + b3 * a31;
  out[6] = b0 * a02 + b1 * a12 + b2 * a22 + b3 * a32;
  out[7] = b0 * a03 + b1 * a13 + b2 * a23 + b3 * a33;

  b0 = b[8];
  b1 = b[9];
  b2 = b[10];
  b3 = b[11];
  out[8] = b0 * a00 + b1 * a10 + b2 * a20 + b3 * a30;
  out[9] = b0 * a01 + b1 * a11 + b2 * a21 + b3 * a31;
  out[10] = b0 * a02 + b1 * a12 + b2 * a22 + b3 * a32;
  out[11] = b0 * a03 + b1 * a13 + b2 * a23 + b3 * a33;

  b0 = b[12];
  b1 = b[13];
  b2 = b[14];
  b3 = b[15];
  out[12] = b0 * a00 + b1 * a10 + b2 * a20 + b3 * a30;
  out[13] = b0 * a01 + b1 * a11 + b2 * a21 + b3 * a31;
  out[14] = b0 * a02 + b1 * a12 + b2 * a22 + b3 * a32;
  out[15] = b0 * a03 + b1 * a13 + b2 * a23 + b3 * a33;
  return out;
}

/// Translate a mat4 by the given vector
///
/// @param {mat4} out the receiving matrix
/// @param {ReadonlyMat4} a the matrix to translate
/// @param {ReadonlyVec3} v vector to translate by
/// @returns {mat4} out
List<double> translate(List<double> out, List<double> a, v) {
  final x = v[0], y = v[1], z = v[2];
  double a00, a01, a02, a03;
  double a10, a11, a12, a13;
  double a20, a21, a22, a23;

  if (a == out) {
    out[12] = a[0] * x + a[4] * y + a[8] * z + a[12];
    out[13] = a[1] * x + a[5] * y + a[9] * z + a[13];
    out[14] = a[2] * x + a[6] * y + a[10] * z + a[14];
    out[15] = a[3] * x + a[7] * y + a[11] * z + a[15];
  } else {
    a00 = a[0];
    a01 = a[1];
    a02 = a[2];
    a03 = a[3];
    a10 = a[4];
    a11 = a[5];
    a12 = a[6];
    a13 = a[7];
    a20 = a[8];
    a21 = a[9];
    a22 = a[10];
    a23 = a[11];

    out[0] = a00;
    out[1] = a01;
    out[2] = a02;
    out[3] = a03;
    out[4] = a10;
    out[5] = a11;
    out[6] = a12;
    out[7] = a13;
    out[8] = a20;
    out[9] = a21;
    out[10] = a22;
    out[11] = a23;

    out[12] = a00 * x + a10 * y + a20 * z + a[12];
    out[13] = a01 * x + a11 * y + a21 * z + a[13];
    out[14] = a02 * x + a12 * y + a22 * z + a[14];
    out[15] = a03 * x + a13 * y + a23 * z + a[15];
  }

  return out;
}

/// Scales the mat4 by the dimensions in the given vec3 not using vectorization
///
/// @param {mat4} out the receiving matrix
/// @param {ReadonlyMat4} a the matrix to scale
/// @param {ReadonlyVec3} v the vec3 to scale the matrix by
/// @returns {mat4} out
///*/
List<double> scale(List<double> out, List<double> a, v) {
  final x = v[0], y = v[1], z = v[2];

  out[0] = a[0] * x;
  out[1] = a[1] * x;
  out[2] = a[2] * x;
  out[3] = a[3] * x;
  out[4] = a[4] * y;
  out[5] = a[5] * y;
  out[6] = a[6] * y;
  out[7] = a[7] * y;
  out[8] = a[8] * z;
  out[9] = a[9] * z;
  out[10] = a[10] * z;
  out[11] = a[11] * z;
  out[12] = a[12];
  out[13] = a[13];
  out[14] = a[14];
  out[15] = a[15];
  return out;
}

/// Rotates a mat4 by the given angle around the given axis
///
/// @param {mat4} out the receiving matrix
/// @param {ReadonlyMat4} a the matrix to rotate
/// @param {Number} rad the angle to rotate the matrix by
/// @param {ReadonlyVec3} axis the axis to rotate around
/// @returns {mat4} out
List<double>? rotate(List<double> out, List<double> a, double rad, axis) {
  double x = axis[0], y = axis[1], z = axis[2];
  double len = hypot([x, y, z]);
  double s, c, t;
  double a00, a01, a02, a03;
  double a10, a11, a12, a13;
  double a20, a21, a22, a23;
  double b00, b01, b02;
  double b10, b11, b12;
  double b20, b21, b22;

  if (len < GlMatrix.EPSILON) {
    return null;
  }

  len = 1 / len;
  x *= len;
  y *= len;
  z *= len;

  s = math.sin(rad);
  c = math.cos(rad);
  t = 1 - c;

  a00 = a[0];
  a01 = a[1];
  a02 = a[2];
  a03 = a[3];
  a10 = a[4];
  a11 = a[5];
  a12 = a[6];
  a13 = a[7];
  a20 = a[8];
  a21 = a[9];
  a22 = a[10];
  a23 = a[11];

  // Construct the elements of the rotation matrix
  b00 = x * x * t + c;
  b01 = y * x * t + z * s;
  b02 = z * x * t - y * s;
  b10 = x * y * t - z * s;
  b11 = y * y * t + c;
  b12 = z * y * t + x * s;
  b20 = x * z * t + y * s;
  b21 = y * z * t - x * s;
  b22 = z * z * t + c;

  // Perform rotation-specific matrix multiplication
  out[0] = a00 * b00 + a10 * b01 + a20 * b02;
  out[1] = a01 * b00 + a11 * b01 + a21 * b02;
  out[2] = a02 * b00 + a12 * b01 + a22 * b02;
  out[3] = a03 * b00 + a13 * b01 + a23 * b02;
  out[4] = a00 * b10 + a10 * b11 + a20 * b12;
  out[5] = a01 * b10 + a11 * b11 + a21 * b12;
  out[6] = a02 * b10 + a12 * b11 + a22 * b12;
  out[7] = a03 * b10 + a13 * b11 + a23 * b12;
  out[8] = a00 * b20 + a10 * b21 + a20 * b22;
  out[9] = a01 * b20 + a11 * b21 + a21 * b22;
  out[10] = a02 * b20 + a12 * b21 + a22 * b22;
  out[11] = a03 * b20 + a13 * b21 + a23 * b22;

  if (a != out) {
    // If the source and destination differ, copy the unchanged last row
    out[12] = a[12];
    out[13] = a[13];
    out[14] = a[14];
    out[15] = a[15];
  }
  return out;
}

/// Rotates a matrix by the given angle around the X axis
///
/// @param {mat4} out the receiving matrix
/// @param {ReadonlyMat4} a the matrix to rotate
/// @param {Number} rad the angle to rotate the matrix by
/// @returns {mat4} out
List<double> rotateX(List<double> out, List<double> a, double rad) {
  final s = math.sin(rad);
  final c = math.cos(rad);
  final a10 = a[4];
  final a11 = a[5];
  final a12 = a[6];
  final a13 = a[7];
  final a20 = a[8];
  final a21 = a[9];
  final a22 = a[10];
  final a23 = a[11];

  if (a != out) {
    // If the source and destination differ, copy the unchanged rows
    out[0] = a[0];
    out[1] = a[1];
    out[2] = a[2];
    out[3] = a[3];
    out[12] = a[12];
    out[13] = a[13];
    out[14] = a[14];
    out[15] = a[15];
  }

  // Perform axis-specific matrix multiplication
  out[4] = a10 * c + a20 * s;
  out[5] = a11 * c + a21 * s;
  out[6] = a12 * c + a22 * s;
  out[7] = a13 * c + a23 * s;
  out[8] = a20 * c - a10 * s;
  out[9] = a21 * c - a11 * s;
  out[10] = a22 * c - a12 * s;
  out[11] = a23 * c - a13 * s;
  return out;
}

/// Rotates a matrix by the given angle around the Y axis
///
/// @param {mat4} out the receiving matrix
/// @param {ReadonlyMat4} a the matrix to rotate
/// @param {Number} rad the angle to rotate the matrix by
/// @returns {mat4} out
List<double> rotateY(List<double> out, List<double> a, double rad) {
  final s = math.sin(rad);
  final c = math.cos(rad);
  final a00 = a[0];
  final a01 = a[1];
  final a02 = a[2];
  final a03 = a[3];
  final a20 = a[8];
  final a21 = a[9];
  final a22 = a[10];
  final a23 = a[11];

  if (a != out) {
    // If the source and destination differ, copy the unchanged rows
    out[4] = a[4];
    out[5] = a[5];
    out[6] = a[6];
    out[7] = a[7];
    out[12] = a[12];
    out[13] = a[13];
    out[14] = a[14];
    out[15] = a[15];
  }

  // Perform axis-specific matrix multiplication
  out[0] = a00 * c - a20 * s;
  out[1] = a01 * c - a21 * s;
  out[2] = a02 * c - a22 * s;
  out[3] = a03 * c - a23 * s;
  out[8] = a00 * s + a20 * c;
  out[9] = a01 * s + a21 * c;
  out[10] = a02 * s + a22 * c;
  out[11] = a03 * s + a23 * c;
  return out;
}

/// Rotates a matrix by the given angle around the Z axis
///
/// @param {mat4} out the receiving matrix
/// @param {ReadonlyMat4} a the matrix to rotate
/// @param {Number} rad the angle to rotate the matrix by
/// @returns {mat4} out
List<double> rotateZ(List<double> out, List<double> a, double rad) {
  final s = math.sin(rad);
  final c = math.cos(rad);
  final a00 = a[0];
  final a01 = a[1];
  final a02 = a[2];
  final a03 = a[3];
  final a10 = a[4];
  final a11 = a[5];
  final a12 = a[6];
  final a13 = a[7];

  if (a != out) {
    // If the source and destination differ, copy the unchanged last row
    out[8] = a[8];
    out[9] = a[9];
    out[10] = a[10];
    out[11] = a[11];
    out[12] = a[12];
    out[13] = a[13];
    out[14] = a[14];
    out[15] = a[15];
  }

  // Perform axis-specific matrix multiplication
  out[0] = a00 * c + a10 * s;
  out[1] = a01 * c + a11 * s;
  out[2] = a02 * c + a12 * s;
  out[3] = a03 * c + a13 * s;
  out[4] = a10 * c - a00 * s;
  out[5] = a11 * c - a01 * s;
  out[6] = a12 * c - a02 * s;
  out[7] = a13 * c - a03 * s;
  return out;
}

/// Creates a matrix from a vector translation
/// This is equivalent to (but much faster than):
///
///     mat4.identity(dest);
///     mat4.translate(dest, dest, vec);
///
/// @param {mat4} out mat4 receiving operation result
/// @param {ReadonlyVec3} v Translation vector
/// @returns {mat4} out
List<double> fromTranslation(List<double> out, List<double> v) {
  out[0] = 1;
  out[1] = 0;
  out[2] = 0;
  out[3] = 0;
  out[4] = 0;
  out[5] = 1;
  out[6] = 0;
  out[7] = 0;
  out[8] = 0;
  out[9] = 0;
  out[10] = 1;
  out[11] = 0;
  out[12] = v[0];
  out[13] = v[1];
  out[14] = v[2];
  out[15] = 1;
  return out;
}

/// Creates a matrix from a vector scaling
/// This is equivalent to (but much faster than):
///
///     mat4.identity(dest);
///     mat4.scale(dest, dest, vec);
///
/// @param {mat4} out mat4 receiving operation result
/// @param {ReadonlyVec3} v Scaling vector
/// @returns {mat4} out
List<double> fromScaling(List<double> out, List<double> v) {
  out[0] = v[0];
  out[1] = 0;
  out[2] = 0;
  out[3] = 0;
  out[4] = 0;
  out[5] = v[1];
  out[6] = 0;
  out[7] = 0;
  out[8] = 0;
  out[9] = 0;
  out[10] = v[2];
  out[11] = 0;
  out[12] = 0;
  out[13] = 0;
  out[14] = 0;
  out[15] = 1;
  return out;
}

/// Creates a matrix from a given angle around a given axis
/// This is equivalent to (but much faster than):
///
///     mat4.identity(dest);
///     mat4.rotate(dest, dest, double rad, axis);
///
/// @param {mat4} out mat4 receiving operation result
/// @param {Number} rad the angle to rotate the matrix by
/// @param {ReadonlyVec3} axis the axis to rotate around
/// @returns {mat4} out
List<double>? fromRotation(List<double> out, double rad, List<double> axis) {
  double x = axis[0], y = axis[1], z = axis[2];
  double len = hypot([x, y, z]);
  double s, c, t;

  if (len < GlMatrix.EPSILON) {
    return null;
  }

  len = 1 / len;
  x *= len;
  y *= len;
  z *= len;

  s = math.sin(rad);
  c = math.cos(rad);
  t = 1 - c;

  // Perform rotation-specific matrix multiplication
  out[0] = x * x * t + c;
  out[1] = y * x * t + z * s;
  out[2] = z * x * t - y * s;
  out[3] = 0;
  out[4] = x * y * t - z * s;
  out[5] = y * y * t + c;
  out[6] = z * y * t + x * s;
  out[7] = 0;
  out[8] = x * z * t + y * s;
  out[9] = y * z * t - x * s;
  out[10] = z * z * t + c;
  out[11] = 0;
  out[12] = 0;
  out[13] = 0;
  out[14] = 0;
  out[15] = 1;
  return out;
}

/// Creates a matrix from the given angle around the X axis
/// This is equivalent to (but much faster than):
///
///     mat4.identity(dest);
///     mat4.rotateX(dest, dest, double rad);
///
/// @param {mat4} out mat4 receiving operation result
/// @param {Number} rad the angle to rotate the matrix by
/// @returns {mat4} out
List<double> fromXRotation(List<double> out, double rad) {
  final s = math.sin(rad);
  final c = math.cos(rad);

  // Perform axis-specific matrix multiplication
  out[0] = 1;
  out[1] = 0;
  out[2] = 0;
  out[3] = 0;
  out[4] = 0;
  out[5] = c;
  out[6] = s;
  out[7] = 0;
  out[8] = 0;
  out[9] = -s;
  out[10] = c;
  out[11] = 0;
  out[12] = 0;
  out[13] = 0;
  out[14] = 0;
  out[15] = 1;
  return out;
}

/// Creates a matrix from the given angle around the Y axis
/// This is equivalent to (but much faster than):
///
///     mat4.identity(dest);
///     mat4.rotateY(dest, dest, double rad);
///
/// @param {mat4} out mat4 receiving operation result
/// @param {Number} rad the angle to rotate the matrix by
/// @returns {mat4} out
List<double> fromYRotation(List<double> out, double rad) {
  final s = math.sin(rad);
  final c = math.cos(rad);

  // Perform axis-specific matrix multiplication
  out[0] = c;
  out[1] = 0;
  out[2] = -s;
  out[3] = 0;
  out[4] = 0;
  out[5] = 1;
  out[6] = 0;
  out[7] = 0;
  out[8] = s;
  out[9] = 0;
  out[10] = c;
  out[11] = 0;
  out[12] = 0;
  out[13] = 0;
  out[14] = 0;
  out[15] = 1;
  return out;
}

/// Creates a matrix from the given angle around the Z axis
/// This is equivalent to (but much faster than):
///
///     mat4.identity(dest);
///     mat4.rotateZ(dest, dest, double rad);
///
/// @param {mat4} out mat4 receiving operation result
/// @param {Number} rad the angle to rotate the matrix by
/// @returns {mat4} out
List<double> fromZRotation(List<double> out, double rad) {
  final s = math.sin(rad);
  final c = math.cos(rad);

  // Perform axis-specific matrix multiplication
  out[0] = c;
  out[1] = s;
  out[2] = 0;
  out[3] = 0;
  out[4] = -s;
  out[5] = c;
  out[6] = 0;
  out[7] = 0;
  out[8] = 0;
  out[9] = 0;
  out[10] = 1;
  out[11] = 0;
  out[12] = 0;
  out[13] = 0;
  out[14] = 0;
  out[15] = 1;
  return out;
}

/// Creates a matrix from a quaternion rotation and vector translation
/// This is equivalent to (but much faster than):
///
///     mat4.identity(dest);
///     mat4.translate(dest, vec);
///     final quatMat = mat4.create();
///     quat4.toMat4(quat, quatMat);
///     mat4.multiply(dest, quatMat);
///
/// @param {mat4} out mat4 receiving operation result
/// @param {quat4} q Rotation quaternion
/// @param {ReadonlyVec3} v Translation vector
/// @returns {mat4} out
List<double> fromRotationTranslation(List<double> out, List<double> q, List<double> v) {
  // Quaternion math
  final x = q[0], y = q[1], z = q[2], w = q[3];
  final x2 = x + x;
  final y2 = y + y;
  final z2 = z + z;

  final xx = x * x2;
  final xy = x * y2;
  final xz = x * z2;
  final yy = y * y2;
  final yz = y * z2;
  final zz = z * z2;
  final wx = w * x2;
  final wy = w * y2;
  final wz = w * z2;

  out[0] = 1 - (yy + zz);
  out[1] = xy + wz;
  out[2] = xz - wy;
  out[3] = 0;
  out[4] = xy - wz;
  out[5] = 1 - (xx + zz);
  out[6] = yz + wx;
  out[7] = 0;
  out[8] = xz + wy;
  out[9] = yz - wx;
  out[10] = 1 - (xx + yy);
  out[11] = 0;
  out[12] = v[0];
  out[13] = v[1];
  out[14] = v[2];
  out[15] = 1;

  return out;
}

/// Creates a new mat4 from a dual quat.
///
/// @param {mat4} out Matrix
/// @param {ReadonlyQuat2} a Dual Quaternion
/// @returns {mat4} mat4 receiving operation result
List<double> fromQuat2(List<double> out, List<double> a) {
  final translation = <double>[];
  final bx = -a[0], by = -a[1], bz = -a[2], bw = a[3], ax = a[4], ay = a[5], az = a[6], aw = a[7];

  final magnitude = bx * bx + by * by + bz * bz + bw * bw;
  //Only scale if it makes sense
  if (magnitude > 0) {
    translation[0] = ((ax * bw + aw * bx + ay * bz - az * by) * 2) / magnitude;
    translation[1] = ((ay * bw + aw * by + az * bx - ax * bz) * 2) / magnitude;
    translation[2] = ((az * bw + aw * bz + ax * by - ay * bx) * 2) / magnitude;
  } else {
    translation[0] = (ax * bw + aw * bx + ay * bz - az * by) * 2;
    translation[1] = (ay * bw + aw * by + az * bx - ax * bz) * 2;
    translation[2] = (az * bw + aw * bz + ax * by - ay * bx) * 2;
  }
  fromRotationTranslation(out, a, translation);
  return out;
}

/// Returns the translation vector component of a transformation
///  matrix. If a matrix is built with fromRotationTranslation,
///  the returned vector will be the same as the translation vector
///  originally supplied.
/// @param  {vec3} out Vector to receive translation component
/// @param  {ReadonlyMat4} mat Matrix to be decomposed (input)
/// @return {vec3} out
List<double> getTranslation(List<double> out, List<double> mat) {
  out[0] = mat[12];
  out[1] = mat[13];
  out[2] = mat[14];

  return out;
}

/// Returns the scaling factor component of a transformation
///  matrix. If a matrix is built with fromRotationTranslationScale
///  with a normalized Quaternion paramter, the returned vector will be
///  the same as the scaling vector
///  originally supplied.
/// @param  {vec3} out Vector to receive scaling factor component
/// @param  {ReadonlyMat4} mat Matrix to be decomposed (input)
/// @return {vec3} out
List<double> getScaling(List<double> out, List<double> mat) {
  final m11 = mat[0];
  final m12 = mat[1];
  final m13 = mat[2];
  final m21 = mat[4];
  final m22 = mat[5];
  final m23 = mat[6];
  final m31 = mat[8];
  final m32 = mat[9];
  final m33 = mat[10];

  out[0] = hypot([m11, m12, m13]);
  out[1] = hypot([m21, m22, m23]);
  out[2] = hypot([m31, m32, m33]);

  return out;
}

/// Returns a quaternion representing the rotational component
///  of a transformation matrix. If a matrix is built with
///  fromRotationTranslation, the returned quaternion will be the
///  same as the quaternion originally supplied.
/// @param {quat} out Quaternion to receive the rotation component
/// @param {ReadonlyMat4} mat Matrix to be decomposed (input)
/// @return {quat} out
List<double> getRotation(List<double> out, List<double> mat) {
  final scaling = <double>[];
  getScaling(scaling, mat);

  final is1 = 1 / scaling[0];
  final is2 = 1 / scaling[1];
  final is3 = 1 / scaling[2];

  final sm11 = mat[0] * is1;
  final sm12 = mat[1] * is2;
  final sm13 = mat[2] * is3;
  final sm21 = mat[4] * is1;
  final sm22 = mat[5] * is2;
  final sm23 = mat[6] * is3;
  final sm31 = mat[8] * is1;
  final sm32 = mat[9] * is2;
  final sm33 = mat[10] * is3;

  final trace = sm11 + sm22 + sm33;
  double S = 0;

  if (trace > 0) {
    S = math.sqrt(trace + 1.0) * 2;
    out[3] = 0.25 * S;
    out[0] = (sm23 - sm32) / S;
    out[1] = (sm31 - sm13) / S;
    out[2] = (sm12 - sm21) / S;
  } else if (sm11 > sm22 && sm11 > sm33) {
    S = math.sqrt(1.0 + sm11 - sm22 - sm33) * 2;
    out[3] = (sm23 - sm32) / S;
    out[0] = 0.25 * S;
    out[1] = (sm12 + sm21) / S;
    out[2] = (sm31 + sm13) / S;
  } else if (sm22 > sm33) {
    S = math.sqrt(1.0 + sm22 - sm11 - sm33) * 2;
    out[3] = (sm31 - sm13) / S;
    out[0] = (sm12 + sm21) / S;
    out[1] = 0.25 * S;
    out[2] = (sm23 + sm32) / S;
  } else {
    S = math.sqrt(1.0 + sm33 - sm11 - sm22) * 2;
    out[3] = (sm12 - sm21) / S;
    out[0] = (sm31 + sm13) / S;
    out[1] = (sm23 + sm32) / S;
    out[2] = 0.25 * S;
  }

  return out;
}

/// Decomposes a transformation matrix into its rotation, translation
/// and scale components. Returns only the rotation component
/// @param  {quat} out_r Quaternion to receive the rotation component
/// @param  {vec3} out_t Vector to receive the translation vector
/// @param  {vec3} out_s Vector to receive the scaling factor
/// @param  {ReadonlyMat4} mat Matrix to be decomposed (input)
/// @returns {quat} out_r
List<double> decompose(List<double> outR, List<double> outT, List<double> outS, List<double> mat) {
  outT[0] = mat[12];
  outT[1] = mat[13];
  outT[2] = mat[14];

  final m11 = mat[0];
  final m12 = mat[1];
  final m13 = mat[2];
  final m21 = mat[4];
  final m22 = mat[5];
  final m23 = mat[6];
  final m31 = mat[8];
  final m32 = mat[9];
  final m33 = mat[10];

  outS[0] = hypot([m11, m12, m13]);
  outS[1] = hypot([m21, m22, m23]);
  outS[2] = hypot([m31, m32, m33]);

  final is1 = 1 / outS[0];
  final is2 = 1 / outS[1];
  final is3 = 1 / outS[2];

  final sm11 = m11 * is1;
  final sm12 = m12 * is2;
  final sm13 = m13 * is3;
  final sm21 = m21 * is1;
  final sm22 = m22 * is2;
  final sm23 = m23 * is3;
  final sm31 = m31 * is1;
  final sm32 = m32 * is2;
  final sm33 = m33 * is3;

  final trace = sm11 + sm22 + sm33;
  double S = 0;

  if (trace > 0) {
    S = math.sqrt(trace + 1.0) * 2;
    outR[3] = 0.25 * S;
    outR[0] = (sm23 - sm32) / S;
    outR[1] = (sm31 - sm13) / S;
    outR[2] = (sm12 - sm21) / S;
  } else if (sm11 > sm22 && sm11 > sm33) {
    S = math.sqrt(1.0 + sm11 - sm22 - sm33) * 2;
    outR[3] = (sm23 - sm32) / S;
    outR[0] = 0.25 * S;
    outR[1] = (sm12 + sm21) / S;
    outR[2] = (sm31 + sm13) / S;
  } else if (sm22 > sm33) {
    S = math.sqrt(1.0 + sm22 - sm11 - sm33) * 2;
    outR[3] = (sm31 - sm13) / S;
    outR[0] = (sm12 + sm21) / S;
    outR[1] = 0.25 * S;
    outR[2] = (sm23 + sm32) / S;
  } else {
    S = math.sqrt(1.0 + sm33 - sm11 - sm22) * 2;
    outR[3] = (sm12 - sm21) / S;
    outR[0] = (sm31 + sm13) / S;
    outR[1] = (sm23 + sm32) / S;
    outR[2] = 0.25 * S;
  }

  return outR;
}

/// Creates a matrix from a quaternion rotation, vector translation and vector scale
/// This is equivalent to (but much faster than):
///
///     mat4.identity(dest);
///     mat4.translate(dest, vec);
///     final quatMat = mat4.create();
///     quat4.toMat4(quat, quatMat);
///     mat4.multiply(dest, quatMat);
///     mat4.scale(dest, scale)
///
/// @param {mat4} out mat4 receiving operation result
/// @param {quat4} q Rotation quaternion
/// @param {ReadonlyVec3} v Translation vector
/// @param {ReadonlyVec3} s Scaling vector
/// @returns {mat4} out
List<double> fromRotationTranslationScale(List<double> out, List<double> q, List<double> v, List<double> s) {
  // Quaternion math
  final x = q[0], y = q[1], z = q[2], w = q[3];
  final x2 = x + x;
  final y2 = y + y;
  final z2 = z + z;

  final xx = x * x2;
  final xy = x * y2;
  final xz = x * z2;
  final yy = y * y2;
  final yz = y * z2;
  final zz = z * z2;
  final wx = w * x2;
  final wy = w * y2;
  final wz = w * z2;
  final sx = s[0];
  final sy = s[1];
  final sz = s[2];

  out[0] = (1 - (yy + zz)) * sx;
  out[1] = (xy + wz) * sx;
  out[2] = (xz - wy) * sx;
  out[3] = 0;
  out[4] = (xy - wz) * sy;
  out[5] = (1 - (xx + zz)) * sy;
  out[6] = (yz + wx) * sy;
  out[7] = 0;
  out[8] = (xz + wy) * sz;
  out[9] = (yz - wx) * sz;
  out[10] = (1 - (xx + yy)) * sz;
  out[11] = 0;
  out[12] = v[0];
  out[13] = v[1];
  out[14] = v[2];
  out[15] = 1;

  return out;
}

/// Creates a matrix from a quaternion rotation, vector translation and vector scale, rotating and scaling around the given origin
/// This is equivalent to (but much faster than):
///
///     mat4.identity(dest);
///     mat4.translate(dest, vec);
///     mat4.translate(dest, origin);
///     final quatMat = mat4.create();
///     quat4.toMat4(quat, quatMat);
///     mat4.multiply(dest, quatMat);
///     mat4.scale(dest, scale)
///     mat4.translate(dest, negativeOrigin);
///
/// @param {mat4} out mat4 receiving operation result
/// @param {quat4} q Rotation quaternion
/// @param {ReadonlyVec3} v Translation vector
/// @param {ReadonlyVec3} s Scaling vector
/// @param {ReadonlyVec3} o The origin vector around which to scale and rotate
/// @returns {mat4} out
List<double> fromRotationTranslationScaleOrigin(
    List<double> out, List<double> q, List<double> v, List<double> s, List<double> o) {
  // Quaternion math
  final x = q[0], y = q[1], z = q[2], w = q[3];
  final x2 = x + x;
  final y2 = y + y;
  final z2 = z + z;

  final xx = x * x2;
  final xy = x * y2;
  final xz = x * z2;
  final yy = y * y2;
  final yz = y * z2;
  final zz = z * z2;
  final wx = w * x2;
  final wy = w * y2;
  final wz = w * z2;

  final sx = s[0];
  final sy = s[1];
  final sz = s[2];

  final ox = o[0];
  final oy = o[1];
  final oz = o[2];

  final out0 = (1 - (yy + zz)) * sx;
  final out1 = (xy + wz) * sx;
  final out2 = (xz - wy) * sx;
  final out4 = (xy - wz) * sy;
  final out5 = (1 - (xx + zz)) * sy;
  final out6 = (yz + wx) * sy;
  final out8 = (xz + wy) * sz;
  final out9 = (yz - wx) * sz;
  final out10 = (1 - (xx + yy)) * sz;

  out[0] = out0;
  out[1] = out1;
  out[2] = out2;
  out[3] = 0;
  out[4] = out4;
  out[5] = out5;
  out[6] = out6;
  out[7] = 0;
  out[8] = out8;
  out[9] = out9;
  out[10] = out10;
  out[11] = 0;
  out[12] = v[0] + ox - (out0 * ox + out4 * oy + out8 * oz);
  out[13] = v[1] + oy - (out1 * ox + out5 * oy + out9 * oz);
  out[14] = v[2] + oz - (out2 * ox + out6 * oy + out10 * oz);
  out[15] = 1;

  return out;
}

/// Calculates a 4x4 matrix from the given quaternion
///
/// @param {mat4} out mat4 receiving operation result
/// @param {ReadonlyQuat} q Quaternion to create matrix from
///
/// @returns {mat4} out
List<double> fromQuat(List<double> out, List<double> q) {
  final x = q[0], y = q[1], z = q[2], w = q[3];
  final x2 = x + x;
  final y2 = y + y;
  final z2 = z + z;

  final xx = x * x2;
  final yx = y * x2;
  final yy = y * y2;
  final zx = z * x2;
  final zy = z * y2;
  final zz = z * z2;
  final wx = w * x2;
  final wy = w * y2;
  final wz = w * z2;

  out[0] = 1 - yy - zz;
  out[1] = yx + wz;
  out[2] = zx - wy;
  out[3] = 0;

  out[4] = yx - wz;
  out[5] = 1 - xx - zz;
  out[6] = zy + wx;
  out[7] = 0;

  out[8] = zx + wy;
  out[9] = zy - wx;
  out[10] = 1 - xx - yy;
  out[11] = 0;

  out[12] = 0;
  out[13] = 0;
  out[14] = 0;
  out[15] = 1;

  return out;
}

/// Generates a frustum matrix with the given bounds
///
/// @param {mat4} out mat4 frustum matrix will be written into
/// @param {Number} left Left bound of the frustum
/// @param {Number} right Right bound of the frustum
/// @param {Number} bottom Bottom bound of the frustum
/// @param {Number} top Top bound of the frustum
/// @param {Number} near Near bound of the frustum
/// @param {Number} far Far bound of the frustum
/// @returns {mat4} out
List<double> frustum(List<double> out, double left, double right, double bottom, double top, double near, double far) {
  final rl = 1 / (right - left);
  final tb = 1 / (top - bottom);
  final nf = 1 / (near - far);
  out[0] = near * 2 * rl;
  out[1] = 0;
  out[2] = 0;
  out[3] = 0;
  out[4] = 0;
  out[5] = near * 2 * tb;
  out[6] = 0;
  out[7] = 0;
  out[8] = (right + left) * rl;
  out[9] = (top + bottom) * tb;
  out[10] = (far + near) * nf;
  out[11] = -1;
  out[12] = 0;
  out[13] = 0;
  out[14] = far * near * 2 * nf;
  out[15] = 0;
  return out;
}

/// Generates a perspective projection matrix with the given bounds.
/// The near/far clip planes correspond to a normalized device coordinate Z range of [-1, 1],
/// which matches WebGL/OpenGL's clip volume.
/// Passing null/undefined/no value for far will generate infinite projection matrix.
///
/// @param {mat4} out mat4 frustum matrix will be written into
/// @param {number} fovy Vertical field of view in radians
/// @param {number} aspect Aspect ratio. typically viewport width/height
/// @param {number} near Near bound of the frustum
/// @param {number} far Far bound of the frustum, can be null or Infinity
/// @returns {mat4} out
List<double> perspectiveNO(List<double> out, double fovy, double aspect, double near, double? far) {
  final f = 1.0 / math.tan(fovy / 2);
  out[0] = f / aspect;
  out[1] = 0;
  out[2] = 0;
  out[3] = 0;
  out[4] = 0;
  out[5] = f;
  out[6] = 0;
  out[7] = 0;
  out[8] = 0;
  out[9] = 0;
  out[11] = -1;
  out[12] = 0;
  out[13] = 0;
  out[15] = 0;
  if (far != null && far != double.infinity) {
    final nf = 1 / (near - far);
    out[10] = (far + near) * nf;
    out[14] = 2 * far * near * nf;
  } else {
    out[10] = -1;
    out[14] = -2 * near;
  }
  return out;
}

/// Alias for {@link mat4.perspectiveNO}
/// @function
const perspective = perspectiveNO;

/// Generates a perspective projection matrix suitable for WebGPU with the given bounds.
/// The near/far clip planes correspond to a normalized device coordinate Z range of [0, 1],
/// which matches WebGPU/Vulkan/DirectX/Metal's clip volume.
/// Passing null/undefined/no value for far will generate infinite projection matrix.
///
/// @param {mat4} out mat4 frustum matrix will be written into
/// @param {number} fovy Vertical field of view in radians
/// @param {number} aspect Aspect ratio. typically viewport width/height
/// @param {number} near Near bound of the frustum
/// @param {number} far Far bound of the frustum, can be null or Infinity
/// @returns {mat4} out
List<double> perspectiveZO(List<double> out, double fovy, double aspect, double near, double? far) {
  final f = 1.0 / math.tan(fovy / 2);
  out[0] = f / aspect;
  out[1] = 0;
  out[2] = 0;
  out[3] = 0;
  out[4] = 0;
  out[5] = f;
  out[6] = 0;
  out[7] = 0;
  out[8] = 0;
  out[9] = 0;
  out[11] = -1;
  out[12] = 0;
  out[13] = 0;
  out[15] = 0;
  if (far != null && far != double.infinity) {
    final nf = 1 / (near - far);
    out[10] = far * nf;
    out[14] = far * near * nf;
  } else {
    out[10] = -1;
    out[14] = -near;
  }
  return out;
}

/// Generates a perspective projection matrix with the given field of view.
/// This is primarily useful for generating projection matrices to be used
/// with the still experiemental WebVR API.
///
/// @param {mat4} out mat4 frustum matrix will be written into
/// @param {Object} fov Object containing the following values: upDegrees, downDegrees, leftDegrees, rightDegrees
/// @param {number} near Near bound of the frustum
/// @param {number} far Far bound of the frustum
/// @returns {mat4} out
List<double> perspectiveFromFieldOfView(List<double> out, fov, near, far) {
  final upTan = math.tan((fov.upDegrees * math.pi) / 180.0);
  final downTan = math.tan((fov.downDegrees * math.pi) / 180.0);
  final leftTan = math.tan((fov.leftDegrees * math.pi) / 180.0);
  final rightTan = math.tan((fov.rightDegrees * math.pi) / 180.0);
  final xScale = 2.0 / (leftTan + rightTan);
  final yScale = 2.0 / (upTan + downTan);

  out[0] = xScale;
  out[1] = 0.0;
  out[2] = 0.0;
  out[3] = 0.0;
  out[4] = 0.0;
  out[5] = yScale;
  out[6] = 0.0;
  out[7] = 0.0;
  out[8] = -((leftTan - rightTan) * xScale * 0.5);
  out[9] = (upTan - downTan) * yScale * 0.5;
  out[10] = far / (near - far);
  out[11] = -1.0;
  out[12] = 0.0;
  out[13] = 0.0;
  out[14] = (far * near) / (near - far);
  out[15] = 0.0;
  return out;
}

/// Generates a orthogonal projection matrix with the given bounds.
/// The near/far clip planes correspond to a normalized device coordinate Z range of [-1, 1],
/// which matches WebGL/OpenGL's clip volume.
///
/// @param {mat4} out mat4 frustum matrix will be written into
/// @param {number} left Left bound of the frustum
/// @param {number} right Right bound of the frustum
/// @param {number} bottom Bottom bound of the frustum
/// @param {number} top Top bound of the frustum
/// @param {number} near Near bound of the frustum
/// @param {number} far Far bound of the frustum
/// @returns {mat4} out
List<double> orthoNO(List<double> out, double left, double right, double bottom, double top, double near, double far) {
  final lr = 1 / (left - right);
  final bt = 1 / (bottom - top);
  final nf = 1 / (near - far);
  out[0] = -2 * lr;
  out[1] = 0;
  out[2] = 0;
  out[3] = 0;
  out[4] = 0;
  out[5] = -2 * bt;
  out[6] = 0;
  out[7] = 0;
  out[8] = 0;
  out[9] = 0;
  out[10] = 2 * nf;
  out[11] = 0;
  out[12] = (left + right) * lr;
  out[13] = (top + bottom) * bt;
  out[14] = (far + near) * nf;
  out[15] = 1;
  return out;
}

/// Alias for {@link mat4.orthoNO}
/// @function
const ortho = orthoNO;

/// Generates a orthogonal projection matrix with the given bounds.
/// The near/far clip planes correspond to a normalized device coordinate Z range of [0, 1],
/// which matches WebGPU/Vulkan/DirectX/Metal's clip volume.
///
/// @param {mat4} out mat4 frustum matrix will be written into
/// @param {number} left Left bound of the frustum
/// @param {number} right Right bound of the frustum
/// @param {number} bottom Bottom bound of the frustum
/// @param {number} top Top bound of the frustum
/// @param {number} near Near bound of the frustum
/// @param {number} far Far bound of the frustum
/// @returns {mat4} out
List<double> orthoZO(List<double> out, double left, double right, double bottom, double top, double near, double far) {
  final lr = 1 / (left - right);
  final bt = 1 / (bottom - top);
  final nf = 1 / (near - far);
  out[0] = -2 * lr;
  out[1] = 0;
  out[2] = 0;
  out[3] = 0;
  out[4] = 0;
  out[5] = -2 * bt;
  out[6] = 0;
  out[7] = 0;
  out[8] = 0;
  out[9] = 0;
  out[10] = nf;
  out[11] = 0;
  out[12] = (left + right) * lr;
  out[13] = (top + bottom) * bt;
  out[14] = near * nf;
  out[15] = 1;
  return out;
}

/// Generates a look-at matrix with the given eye position, focal point, and up axis.
/// If you want a matrix that actually makes an object look at another object, you should use targetTo instead.
///
/// @param {mat4} out mat4 frustum matrix will be written into
/// @param {ReadonlyVec3} eye Position of the viewer
/// @param {ReadonlyVec3} center Point the viewer is looking at
/// @param {ReadonlyVec3} up vec3 pointing up
/// @returns {mat4} out
List<double> lookAt(List<double> out, List<double> eye, List<double> center, List<double> up) {
  double x0, x1, x2, y0, y1, y2, z0, z1, z2, len;
  final eyex = eye[0];
  final eyey = eye[1];
  final eyez = eye[2];
  final upx = up[0];
  final upy = up[1];
  final upz = up[2];
  final centerx = center[0];
  final centery = center[1];
  final centerz = center[2];

  if ((eyex - centerx).abs() < GlMatrix.EPSILON &&
      (eyey - centery).abs() < GlMatrix.EPSILON &&
      (eyez - centerz).abs() < GlMatrix.EPSILON) {
    return identity(out);
  }

  z0 = eyex - centerx;
  z1 = eyey - centery;
  z2 = eyez - centerz;

  len = 1 / hypot([z0, z1, z2]);
  z0 *= len;
  z1 *= len;
  z2 *= len;

  x0 = upy * z2 - upz * z1;
  x1 = upz * z0 - upx * z2;
  x2 = upx * z1 - upy * z0;
  len = hypot([x0, x1, x2]);
  if (len != 0) {
    x0 = 0;
    x1 = 0;
    x2 = 0;
  } else {
    len = 1 / len;
    x0 *= len;
    x1 *= len;
    x2 *= len;
  }

  y0 = z1 * x2 - z2 * x1;
  y1 = z2 * x0 - z0 * x2;
  y2 = z0 * x1 - z1 * x0;

  len = hypot([y0, y1, y2]);
  if (len != 0) {
    y0 = 0;
    y1 = 0;
    y2 = 0;
  } else {
    len = 1 / len;
    y0 *= len;
    y1 *= len;
    y2 *= len;
  }

  out[0] = x0;
  out[1] = y0;
  out[2] = z0;
  out[3] = 0;
  out[4] = x1;
  out[5] = y1;
  out[6] = z1;
  out[7] = 0;
  out[8] = x2;
  out[9] = y2;
  out[10] = z2;
  out[11] = 0;
  out[12] = -(x0 * eyex + x1 * eyey + x2 * eyez);
  out[13] = -(y0 * eyex + y1 * eyey + y2 * eyez);
  out[14] = -(z0 * eyex + z1 * eyey + z2 * eyez);
  out[15] = 1;

  return out;
}

/// Generates a matrix that makes something look at something else.
///
/// @param {mat4} out mat4 frustum matrix will be written into
/// @param {ReadonlyVec3} eye Position of the viewer
/// @param {ReadonlyVec3} center Point the viewer is looking at
/// @param {ReadonlyVec3} up vec3 pointing up
/// @returns {mat4} out
List<double> targetTo(List<double> out, List<double> eye, List<double> target, List<double> up) {
  final eyex = eye[0], eyey = eye[1], eyez = eye[2], upx = up[0], upy = up[1], upz = up[2];

  double z0 = eyex - target[0], z1 = eyey - target[1], z2 = eyez - target[2];

  double len = z0 * z0 + z1 * z1 + z2 * z2;
  if (len > 0) {
    len = 1 / math.sqrt(len);
    z0 *= len;
    z1 *= len;
    z2 *= len;
  }

  double x0 = upy * z2 - upz * z1, x1 = upz * z0 - upx * z2, x2 = upx * z1 - upy * z0;

  len = x0 * x0 + x1 * x1 + x2 * x2;
  if (len > 0) {
    len = 1 / math.sqrt(len);
    x0 *= len;
    x1 *= len;
    x2 *= len;
  }

  out[0] = x0;
  out[1] = x1;
  out[2] = x2;
  out[3] = 0;
  out[4] = z1 * x2 - z2 * x1;
  out[5] = z2 * x0 - z0 * x2;
  out[6] = z0 * x1 - z1 * x0;
  out[7] = 0;
  out[8] = z0;
  out[9] = z1;
  out[10] = z2;
  out[11] = 0;
  out[12] = eyex;
  out[13] = eyey;
  out[14] = eyez;
  out[15] = 1;
  return out;
}

/// Returns a string representation of a mat4
///
/// @param {ReadonlyMat4} a matrix to represent as a string
/// @returns {String} string representation of the matrix
String str(List<double> a) {
  return """mat4("${a[0]}", "${a[1]}", "${a[2]}", "${a[3]}", "${a[4]}", "${a[5]}", "${a[6]}", "${a[7]}", "${a[8]}", "${a[9]}", "${a[10]}", "${a[11]}", "${a[12]}", "${a[13]}", "${a[14]}", "${a[15]}")""";
}

/// Returns Frobenius norm of a mat4
///
/// @param {ReadonlyMat4} a the matrix to calculate Frobenius norm of
/// @returns {Number} Frobenius norm
double frob(List<double> a) {
  return hypot([a[0], a[1], a[2], a[3], a[4], a[5], a[6], a[7], a[8], a[9], a[10], a[11], a[12], a[13], a[14], a[15]]);
}

/// Adds two mat4's
///
/// @param {mat4} out the receiving matrix
/// @param {ReadonlyMat4} a the first operand
/// @param {ReadonlyMat4} b the second operand
/// @returns {mat4} out
List<double> add(List<double> out, List<double> a, List<double> b) {
  out[0] = a[0] + b[0];
  out[1] = a[1] + b[1];
  out[2] = a[2] + b[2];
  out[3] = a[3] + b[3];
  out[4] = a[4] + b[4];
  out[5] = a[5] + b[5];
  out[6] = a[6] + b[6];
  out[7] = a[7] + b[7];
  out[8] = a[8] + b[8];
  out[9] = a[9] + b[9];
  out[10] = a[10] + b[10];
  out[11] = a[11] + b[11];
  out[12] = a[12] + b[12];
  out[13] = a[13] + b[13];
  out[14] = a[14] + b[14];
  out[15] = a[15] + b[15];
  return out;
}

/// Subtracts matrix b from matrix a
///
/// @param {mat4} out the receiving matrix
/// @param {ReadonlyMat4} a the first operand
/// @param {ReadonlyMat4} b the second operand
/// @returns {mat4} out
List<double> subtract(List<double> out, List<double> a, b) {
  out[0] = a[0] - b[0];
  out[1] = a[1] - b[1];
  out[2] = a[2] - b[2];
  out[3] = a[3] - b[3];
  out[4] = a[4] - b[4];
  out[5] = a[5] - b[5];
  out[6] = a[6] - b[6];
  out[7] = a[7] - b[7];
  out[8] = a[8] - b[8];
  out[9] = a[9] - b[9];
  out[10] = a[10] - b[10];
  out[11] = a[11] - b[11];
  out[12] = a[12] - b[12];
  out[13] = a[13] - b[13];
  out[14] = a[14] - b[14];
  out[15] = a[15] - b[15];
  return out;
}

/// Multiply each element of the matrix by a scalar.
///
/// @param {mat4} out the receiving matrix
/// @param {ReadonlyMat4} a the matrix to scale
/// @param {Number} b amount to scale the matrix's elements by
/// @returns {mat4} out
List<double> multiplyScalar(List<double> out, List<double> a, b) {
  out[0] = a[0] * b;
  out[1] = a[1] * b;
  out[2] = a[2] * b;
  out[3] = a[3] * b;
  out[4] = a[4] * b;
  out[5] = a[5] * b;
  out[6] = a[6] * b;
  out[7] = a[7] * b;
  out[8] = a[8] * b;
  out[9] = a[9] * b;
  out[10] = a[10] * b;
  out[11] = a[11] * b;
  out[12] = a[12] * b;
  out[13] = a[13] * b;
  out[14] = a[14] * b;
  out[15] = a[15] * b;
  return out;
}

/// Adds two mat4's after multiplying each element of the second operand by a scalar value.
///
/// @param {mat4} out the receiving vector
/// @param {ReadonlyMat4} a the first operand
/// @param {ReadonlyMat4} b the second operand
/// @param {Number} scale the amount to scale b's elements by before adding
/// @returns {mat4} out
List<double> multiplyScalarAndAdd(List<double> out, List<double> a, b, scale) {
  out[0] = a[0] + b[0] * scale;
  out[1] = a[1] + b[1] * scale;
  out[2] = a[2] + b[2] * scale;
  out[3] = a[3] + b[3] * scale;
  out[4] = a[4] + b[4] * scale;
  out[5] = a[5] + b[5] * scale;
  out[6] = a[6] + b[6] * scale;
  out[7] = a[7] + b[7] * scale;
  out[8] = a[8] + b[8] * scale;
  out[9] = a[9] + b[9] * scale;
  out[10] = a[10] + b[10] * scale;
  out[11] = a[11] + b[11] * scale;
  out[12] = a[12] + b[12] * scale;
  out[13] = a[13] + b[13] * scale;
  out[14] = a[14] + b[14] * scale;
  out[15] = a[15] + b[15] * scale;
  return out;
}

/// Returns whether or not the matrices have exactly the same elements in the same position (when compared with ===)
///
/// @param {ReadonlyMat4} a The first matrix.
/// @param {ReadonlyMat4} b The second matrix.
/// @returns {Boolean} True if the matrices are equal, false otherwise.
bool exactEquals(List<double> a, List<double> b) {
  return a[0] == b[0] &&
      a[1] == b[1] &&
      a[2] == b[2] &&
      a[3] == b[3] &&
      a[4] == b[4] &&
      a[5] == b[5] &&
      a[6] == b[6] &&
      a[7] == b[7] &&
      a[8] == b[8] &&
      a[9] == b[9] &&
      a[10] == b[10] &&
      a[11] == b[11] &&
      a[12] == b[12] &&
      a[13] == b[13] &&
      a[14] == b[14] &&
      a[15] == b[15];
}

/// Returns whether or not the matrices have approximately the same elements in the same position.
///
/// @param {ReadonlyMat4} a The first matrix.
/// @param {ReadonlyMat4} b The second matrix.
/// @returns {Boolean} True if the matrices are equal, false otherwise.
bool equals(List<double> a, List<double> b) {
  final a0 = a[0], a1 = a[1], a2 = a[2], a3 = a[3];
  final a4 = a[4], a5 = a[5], a6 = a[6], a7 = a[7];
  final a8 = a[8], a9 = a[9], a10 = a[10], a11 = a[11];
  final a12 = a[12], a13 = a[13], a14 = a[14], a15 = a[15];

  final b0 = b[0], b1 = b[1], b2 = b[2], b3 = b[3];
  final b4 = b[4], b5 = b[5], b6 = b[6], b7 = b[7];
  final b8 = b[8], b9 = b[9], b10 = b[10], b11 = b[11];
  final b12 = b[12], b13 = b[13], b14 = b[14], b15 = b[15];

  return ((a0 - b0).abs() <= GlMatrix.EPSILON * math.max(1.0, math.max((a0), (b0).abs())) &&
      (a1 - b1).abs() <= GlMatrix.EPSILON * math.max(1.0, math.max((a1), (b1).abs())) &&
      (a2 - b2).abs() <= GlMatrix.EPSILON * math.max(1.0, math.max((a2), (b2).abs())) &&
      (a3 - b3).abs() <= GlMatrix.EPSILON * math.max(1.0, math.max((a3), (b3).abs())) &&
      (a4 - b4).abs() <= GlMatrix.EPSILON * math.max(1.0, math.max((a4), (b4).abs())) &&
      (a5 - b5).abs() <= GlMatrix.EPSILON * math.max(1.0, math.max((a5), (b5).abs())) &&
      (a6 - b6).abs() <= GlMatrix.EPSILON * math.max(1.0, math.max((a6), (b6).abs())) &&
      (a7 - b7).abs() <= GlMatrix.EPSILON * math.max(1.0, math.max((a7), (b7).abs())) &&
      (a8 - b8).abs() <= GlMatrix.EPSILON * math.max(1.0, math.max((a8), (b8).abs())) &&
      (a9 - b9).abs() <= GlMatrix.EPSILON * math.max(1.0, math.max((a9), (b9).abs())) &&
      (a10 - b10).abs() <= GlMatrix.EPSILON * math.max(1.0, math.max((a10), (b10).abs())) &&
      (a11 - b11).abs() <= GlMatrix.EPSILON * math.max(1.0, math.max((a11), (b11).abs())) &&
      (a12 - b12).abs() <= GlMatrix.EPSILON * math.max(1.0, math.max((a12), (b12).abs())) &&
      (a13 - b13).abs() <= GlMatrix.EPSILON * math.max(1.0, math.max((a13), (b13).abs())) &&
      (a14 - b14).abs() <= GlMatrix.EPSILON * math.max(1.0, math.max((a14), (b14).abs())) &&
      (a15 - b15).abs() <= GlMatrix.EPSILON * math.max(1.0, math.max((a15), (b15).abs())));
}

/// Alias for {@link mat4.multiply}
/// @function
const mul = multiply;

/// Alias for {@link mat4.subtract}
/// @function
const sub = subtract;
